@@ -29,6 +29,7 @@ enum {
 enum {
     PAGE_ADDED,
     PAGE_REMOVED,
+    REORDERED,
     CLEARED,
     LAST_SIGNAL
 };
@@ -91,6 +92,18 @@ book_append_page (Book *book, gint width, gint height, gint dpi, ScanDirection s
 }
 
 
+void
+book_move_page (Book *book, Page *page, gint location)
+{
+    book->priv->pages = g_list_remove (book->priv->pages, page);
+    book->priv->pages = g_list_insert (book->priv->pages, page, location);
+
+    g_signal_emit (book, signals[REORDERED], 0, page);
+
+    book_set_needs_saving (book, TRUE);
+}
+
+
 void
 book_delete_page (Book *book, Page *page)
 {
@@ -121,6 +134,13 @@ book_get_page (Book *book, gint page_number)
 }
 
 
+gint
+book_get_page_index (Book *book, Page *page)
+{
+     return g_list_index (book->priv->pages, page);
+}
+
+
 static GFile *
 make_indexed_file (const gchar *uri, gint i)
 {
@@ -434,17 +454,17 @@ book_save_pdf (Book *book, GFile *file, GError **error)
         Page *page;
         GdkPixbuf *image;
         guchar *pixels, *data, *compressed_data;
-        gchar *command;
+        gchar *command, width_buffer[G_ASCII_DTOSTR_BUF_SIZE], height_buffer[G_ASCII_DTOSTR_BUF_SIZE];
         const gchar *color_space, *filter = NULL;
         float page_width, page_height;
 
         page = book_get_page (book, i);
-        width = page_get_width (page);
-        height = page_get_height (page);
+        image = page_get_image (page, TRUE);
+        width = gdk_pixbuf_get_width (image);
+        height = gdk_pixbuf_get_height (image);
+        pixels = gdk_pixbuf_get_pixels (image);
         page_width = width * 72. / page_get_dpi (page);
         page_height = height * 72. / page_get_dpi (page);
-        image = page_get_image (page, TRUE);
-        pixels = gdk_pixbuf_get_pixels (image);
 
         if (page_is_color (page)) {
             int row;
@@ -600,7 +620,9 @@ book_save_pdf (Book *book, GFile *file, GError **error)
         pdf_printf (writer, "/Type /Page\n");
         pdf_printf (writer, "/Parent %d 0 R\n", pages_number);
         pdf_printf (writer, "/Resources << /XObject << /Im%d %d 0 R >> >>\n", i, number+1);
-        pdf_printf (writer, "/MediaBox [ 0 0 %.2f %.2f ]\n", page_width, page_height);
+        pdf_printf (writer, "/MediaBox [ 0 0 %s %s ]\n",
+                    g_ascii_formatd (width_buffer, sizeof (width_buffer), "%.2f", page_width),
+                    g_ascii_formatd (height_buffer, sizeof (height_buffer), "%.2f", page_height));
         pdf_printf (writer, "/Contents %d 0 R\n", number+2);
         pdf_printf (writer, ">>\n");
         pdf_printf (writer, "endobj\n");
@@ -629,9 +651,12 @@ book_save_pdf (Book *book, GFile *file, GError **error)
 
         /* Page contents */
         command = g_strdup_printf ("q\n"
-                                   "%f 0 0 %f 0 0 cm\n"
+                                   "%s 0 0 %s 0 0 cm\n"
                                    "/Im%d Do\n"
-                                   "Q", page_width, page_height, i);
+                                   "Q",
+                                   g_ascii_formatd (width_buffer, sizeof (width_buffer), "%f", page_width),
+                                   g_ascii_formatd (height_buffer, sizeof (height_buffer), "%f", page_height),
+                                   i);
         pdf_printf (writer, "\n");
         number = pdf_start_object (writer);
         pdf_printf (writer, "%d 0 obj\n", number);
@@ -805,6 +830,14 @@ book_class_init (BookClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__OBJECT,
                       G_TYPE_NONE, 1, page_get_type ());
+    signals[REORDERED] =
+        g_signal_new ("reordered",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (BookClass, reordered),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
     signals[CLEARED] =
         g_signal_new ("cleared",
                       G_TYPE_FROM_CLASS (klass),
