@@ -792,7 +792,7 @@ close_device (Scanner *scanner)
         g_debug ("sane_close ()");
         scanner->priv->handle = NULL;
     }
-  
+
     g_free (scanner->priv->buffer);
     scanner->priv->buffer = NULL;
 
@@ -1081,6 +1081,7 @@ do_get_option (Scanner *scanner)
             "Gray",
             "Grayscale",
             SANE_I18N ("Grayscale"),
+            "True Gray", /* Seen in the proprietary brother3 driver */
             NULL
         };
         const char *lineart_scan_modes[] =
@@ -1096,8 +1097,9 @@ do_get_option (Scanner *scanner)
             "Thresholded",
             SANE_VALUE_SCAN_MODE_GRAY,
             "Gray",
-             "Grayscale",
+            "Grayscale",
             SANE_I18N ("Grayscale"),
+            "True Gray", /* Seen in the proprietary brother3 driver */
             NULL
         };
             
@@ -1160,7 +1162,7 @@ do_get_option (Scanner *scanner)
         }
         else if (strcmp (option->name, "three-pass") == 0) {
             set_bool_option (scanner->priv->handle, option, option_index, FALSE, NULL);
-        }                    
+        }
         else if (strcmp (option->name, "test-picture") == 0) {
             set_string_option (scanner->priv->handle, option, option_index, "Color pattern", NULL);
         }
@@ -1255,6 +1257,11 @@ do_get_parameters (Scanner *scanner)
     info->width = scanner->priv->parameters.pixels_per_line;
     info->height = scanner->priv->parameters.lines;
     info->depth = scanner->priv->parameters.depth;
+    /* Reduce bit depth if requested lower than received */
+    // FIXME: This a hack and only works on 8 bit gray to 2 bit gray
+    if (scanner->priv->parameters.depth == 8 && scanner->priv->parameters.format == SANE_FRAME_GRAY && job->depth == 2 && job->scan_mode == SCAN_MODE_GRAY)
+        info->depth = job->depth;
+    info->n_channels = scanner->priv->parameters.format == SANE_FRAME_GRAY ? 1 : 3;
     info->dpi = job->dpi; // FIXME: This is the requested DPI, not the actual DPI
     info->device = g_strdup (scanner->priv->current_device);
 
@@ -1308,10 +1315,13 @@ do_complete_page (Scanner *scanner)
 static void
 do_read (Scanner *scanner)
 {
+    ScanJob *job;
     SANE_Status status;
     SANE_Int n_to_read, n_read;
     gboolean full_read = FALSE;
 
+    job = (ScanJob *) scanner->priv->job_queue->data;
+
     /* Read as many bytes as we expect */
     n_to_read = scanner->priv->buffer_size - scanner->priv->n_used;
 
@@ -1320,10 +1330,12 @@ do_read (Scanner *scanner)
                         n_to_read, &n_read);
     g_debug ("sane_read (%d) -> (%s, %d)", n_to_read, get_status_string (status), n_read);
 
-    /* End of variable length frame */
-    if (status == SANE_STATUS_EOF &&
-        scanner->priv->parameters.lines == -1 &&
-        scanner->priv->bytes_remaining == scanner->priv->parameters.bytes_per_line) {
+    /* Completed read */
+    if (status == SANE_STATUS_EOF) {
+        if (scanner->priv->parameters.lines > 0 && scanner->priv->line_count != scanner->priv->parameters.lines)
+            g_warning ("Scan completed with %d lines, expected %d lines", scanner->priv->parameters.lines, scanner->priv->parameters.lines);
+        if (scanner->priv->n_used > 0)
+            g_warning ("Scan complete with %d bytes of unused data", scanner->priv->n_used);
         do_complete_page (scanner);
         return;
     }
@@ -1344,23 +1356,24 @@ do_read (Scanner *scanner)
     /* Feed out lines */
     if (scanner->priv->n_used >= scanner->priv->parameters.bytes_per_line) {
         ScanLine *line;
+        int i, n_remaining;
 
         line = g_malloc(sizeof(ScanLine));
         switch (scanner->priv->parameters.format) {
         case SANE_FRAME_GRAY:
-            line->format = LINE_GRAY;
+            line->channel = 0;
             break;
         case SANE_FRAME_RGB:
-            line->format = LINE_RGB;
+            line->channel = -1;
             break;
         case SANE_FRAME_RED:
-            line->format = LINE_RED;
+            line->channel = 0;
             break;
         case SANE_FRAME_GREEN:
-            line->format = LINE_GREEN;
+            line->channel = 1;
             break;
         case SANE_FRAME_BLUE:
-            line->format = LINE_BLUE;
+            line->channel = 2;
             break;
         }
         line->width = scanner->priv->parameters.pixels_per_line;
@@ -1373,27 +1386,68 @@ do_read (Scanner *scanner)
         scanner->priv->buffer = NULL;
         scanner->priv->line_count += line->n_lines;
 
-        /* On last line */
-        if (scanner->priv->parameters.lines > 0 && scanner->priv->line_count >= scanner->priv->parameters.lines) {
-            emit_signal (scanner, GOT_LINE, line);
-            do_complete_page (scanner);
+        /* Increase buffer size if did full read */
+        if (full_read)
+            scanner->priv->buffer_size += scanner->priv->parameters.bytes_per_line;
+
+        scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * scanner->priv->buffer_size);
+        n_remaining = scanner->priv->n_used - (line->n_lines * line->data_length);
+        scanner->priv->n_used = 0;
+        for (i = 0; i < n_remaining; i++) {
+            scanner->priv->buffer[i] = line->data[i + (line->n_lines * line->data_length)];
+            scanner->priv->n_used++;
         }
-        else {
-            int i, n_remaining;
 
-            /* Increase buffer size if did full read */
-            if (full_read)
-                scanner->priv->buffer_size += scanner->priv->parameters.bytes_per_line;
+        /* Reduce bit depth if requested lower than received */
+        // FIXME: This a hack and only works on 8 bit gray to 2 bit gray
+        if (scanner->priv->parameters.depth == 8 && scanner->priv->parameters.format == SANE_FRAME_GRAY &&
+            job->depth == 2 && job->scan_mode == SCAN_MODE_GRAY) {
+            gint block_shift = 6;
+            guchar block = 0;
+            guchar *write_ptr = line->data;
 
-            scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * scanner->priv->buffer_size);
-            n_remaining = scanner->priv->n_used - (line->n_lines * line->data_length);
-            scanner->priv->n_used = 0;
-            for (i = 0; i < n_remaining; i++) {
-                scanner->priv->buffer[i] = line->data[i + (line->n_lines * line->data_length)];
-                scanner->priv->n_used++;                
+            for (i = 0; i < line->n_lines; i++) {
+                guchar *in_line = line->data + i * line->data_length;
+                gint x;
+
+                for (x = 0; x < line->width; x++) {
+                     guchar *p = in_line + x;
+                     guchar sample;
+
+                     if (p[0] >= 192)
+                         sample = 3;
+                     else if (p[0] >= 128)
+                         sample = 2;
+                     else if (p[0] >= 64)
+                         sample = 1;
+                     else
+                         sample = 0;
+
+                     block |= sample << block_shift;
+                     if (block_shift == 0) {
+                         *write_ptr = block;
+                         write_ptr++;
+                         block = 0;
+                         block_shift = 6;
+                     }
+                     else {
+                         block_shift -= 2;
+                     }
+                }
+
+                /* Finish each line on a byte boundary */
+                if (block_shift != 6)  {
+                    *write_ptr = block;
+                    write_ptr++;
+                    block = 0;
+                    block_shift = 6;
+                }
             }
-            emit_signal (scanner, GOT_LINE, line);
+
+            line->data_length = (line->width * 2 + 7) / 8;
         }
+
+        emit_signal (scanner, GOT_LINE, line);
     }
 }
 
